/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain.comm;

import java.time.Duration;
import java.util.HashSet;
import java.util.Set;

import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.PrognoseIntervall;

/**
 * Anfrage zum Anfordern einer Prognose-Ganglinie von der Ganglinienprognose.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class PrognoseAnfrage {

    /** Messquerschnitt für den eine Ganglinie angefragt wird. */
    private MessQuerschnitt messQuerschnitt;

    /** Der Zeitraum in für den die Ganglinie bestimmt werden soll. */
    private PrognoseIntervall prognoseZeitraum;

    /** Nur Auswahlverfahren der langfristigen Prognose benutzen? */
    private boolean nurLangfristigeAuswahl;

    /** Diese Ereignistypen werden bei der Ganglinienauswahl ignoriert. */
    private final Set<EreignisTyp> ereignisTypen = new HashSet<>();

    /** Soll eine zyklische Prognose erstellt werden? */
    private boolean zyklischePrognose;

    /** Spätestens nach dieser Zeit die Prognose prüfen. */
    private Duration pruefIntervall;

    /** Maximale &Auml;nderung in Prozent zwischen zwei zyklischen Prognosen. */
    private int schwelle;

    /** Spätestens nach dieser Zeit eine Prognose publizieren. */
    private Duration sendeIntervall;

    /**
     * Generiert eine einmalige Anfrage. Die Option "zyklische Anfrage" wird auf
     * {@code false} gesetzt und die davon abhängigen Parameter mit Defaultwerten
     * belegt.
     *
     * @param mq                     der Messquerschnitt für den eine Ganglinie
     *                               angefragt wird.
     * @param prognoseZeitraum       der Zeitraum der Prognose.
     * @param nurLangfristigeAuswahl Nur Auswahlverfahren der langfristigen Prognose
     *                               benutzen?
     */
    public PrognoseAnfrage(final MessQuerschnitt mq, final PrognoseIntervall prognoseZeitraum,
            final boolean nurLangfristigeAuswahl) {
        this(mq, prognoseZeitraum, nurLangfristigeAuswahl, false, Duration.ofSeconds(1), 0,
                Duration.ofSeconds(1));
    }

    /**
     * Generiert eine Anfrage.
     *
     * @param mq                     der Messquerschnitt für den eine Ganglinie
     *                               angefragt wird.
     * @param prognoseZeitraum       der Zeitraum der Prognose.
     * @param nurLangfristigeAuswahl Nur Auswahlverfahren der langfristigen Prognose
     *                               benutzen?
     * @param zyklischePrognose      Soll eine zyklische Prognose erstellt werden?
     * @param pruefIntervall         Spätestens nach dieser Zeit die Prognose
     *                               prüfen.
     * @param schwelle               Maximale &Auml;nderung in Prozent zwischen zwei
     *                               zyklischen Prognosen.
     * @param sendeIntervall         Spätestens nach dieser Zeit eine Prognose
     *                               publizieren.
     */
    public PrognoseAnfrage(final MessQuerschnitt mq, final PrognoseIntervall prognoseZeitraum,
            final boolean nurLangfristigeAuswahl, final boolean zyklischePrognose, final Duration pruefIntervall,
            final int schwelle, final Duration sendeIntervall) {

        messQuerschnitt = mq;
        this.prognoseZeitraum = prognoseZeitraum;
        this.nurLangfristigeAuswahl = nurLangfristigeAuswahl;
        this.zyklischePrognose = zyklischePrognose;
        this.pruefIntervall = pruefIntervall;
        this.schwelle = schwelle;
        this.sendeIntervall = sendeIntervall;
    }

    /**
     * Baut aus den Informationen der Anfrage ein Datum.
     * <p>
     * Hinweis: Das Ergebnis wird auch im Parameter abgelegt!
     * <p>
     * <em>Hinweis:</em> Diese Methode ist nicht Teil der öffentlichen API und
     * sollte nicht außerhalb der Ganglinie-API verwendet werden.
     *
     * @param daten ein Datum, welches eine (leere) Anfrage darstellt.
     * @return das ausgefüllte Datum.
     */

    /**
     * Gibt die Menge der ausgeschlossenen Ereignistypen zurück.
     *
     * @return Ereignistypeniterator
     */
    public Set<EreignisTyp> getEreignisTypen() {
        return new HashSet<>(ereignisTypen);
    }

     /**
     * Gibt den Messquerschnitt für den eine Ganglinie angefragt wird zurück.
     *
     * @return Ein Messquerschnitt
     */
    public MessQuerschnitt getMessQuerschnitt() {
        return messQuerschnitt;
    }

    /**
     * Gibt den Prognosezeitraum zurück.
     *
     * @return der Zeitraum für den die Ganglinie bestimmt wird oder {@code null},
     *         wenn kein Intervall gesetzt ist bzw. das Intervall ungültig ist. Bei
     *         einem ungültigen Intervall liegt der Startzeitpunkt <em>hinter</em>
     *         dem Endzeitpunkt.
     */
    public PrognoseIntervall getPrognoseZeitraum() {
        return prognoseZeitraum;
    }

    /**
     * Spätestens nach dieser Zeit wird die Prognose geprüft.
     *
     * @return das Prüfintervall
     * @see #isZyklischePrognose()
     */
    public Duration getPruefIntervall() {
        return pruefIntervall;
    }

    /**
     * Maximale &Auml;nderung in Prozent zwischen zwei zyklischen Prognosen. Wird
     * dieser Schwellwert überschritten, wird eine neue Prognose publiziert.
     *
     * @return Schwellwert in Prozent.
     * @see #isZyklischePrognose()
     */
    public int getSchwelle() {
        return schwelle;
    }

    /**
     * Spätestens nach dieser Zeit wird eine Prognose publiziert. Die Ganglinie wird
     * nach dieser Zeit auch publiziert, wenn sie sich nicht geändert hat.
     *
     * @return der Zyklus des Publizierens.
     * @see #isZyklischePrognose()
     */
    public Duration getSendeIntervall() {
        return sendeIntervall;
    }

    /**
     * Sollen nur Auswahlverfahren der langfristigen Prognose benutzt werden?
     *
     * @return {@code true}, wenn dies der Fall ist, sonst {@code false}.
     */
    public boolean isNurLangfristigeAuswahl() {
        return nurLangfristigeAuswahl;
    }

    /**
     * Gibt zurück, ob es sich um eine zyklische oder einmalige Prognose handelt.
     *
     * @return {@code true}, wenn die Prognose zyklisch wiederholt wird und
     *         {@code false}, wenn die Prognose nur einmal durchgeführt wird
     * @see #getPruefIntervall()
     * @see #getSchwelle()
     * @see #getSendeIntervall()
     */
    public boolean isZyklischePrognose() {
        return zyklischePrognose;
    }

    /**
     * Fügt der Filterliste einen Ereignistyp hinzu.
     *
     * @param typ ein Ereignistyp der bei der Ganglinienprognose ignoriert werden
     *            soll.
     */
    public void addEreignisTyp(final EreignisTyp typ) {
        ereignisTypen.add(typ);
    }

    /**
     * Entfernt einen Ereignistyp aus der Filterliste.
     *
     * @param typ ein Ereignistyp der bei der Ganglinienprognose ignoriert werden
     *            soll.
     * @return {@code true}, wenn der Typ enthalten war und {@code false}, wenn er
     *         bereits enthalten war.
     */
    public boolean removeEreignisTyp(final EreignisTyp typ) {
        return ereignisTypen.remove(typ);
    }

    /**
     * Ändert den Messquerschnitt für den angefragt wird.
     *
     * @param messQuerschnitt ein Messquerschnitt.
     */
    public void setMessQuerschnitt(final MessQuerschnitt messQuerschnitt) {
        this.messQuerschnitt = messQuerschnitt;
    }

    /**
     * Ändert den Prognosezeitraum.
     *
     * @param prognoseZeitraum der neue Prognosezeitraum.
     */
    public void setPrognoseZeitraum(final PrognoseIntervall prognoseZeitraum) {
        this.prognoseZeitraum = prognoseZeitraum;
    }

    /**
     * Setzt das Flag für die langfristigen Auswahlmethoden.
     *
     * @param nurLangfristigeAuswahl {@code true}, wenn nur langfristige
     *                               Auswahlmethoden verwendet werden sollen.
     */
    public void setNurLangfristigeAuswahl(final boolean nurLangfristigeAuswahl) {
        this.nurLangfristigeAuswahl = nurLangfristigeAuswahl;
    }

    /**
     * Setzt Flag für eine zyklische Prognose.
     *
     * @param zyklischePrognose {@code false}, wenn die Prognose einmalig ausgeführt
     *                          werden soll und {@code true}, wenn die Prognose
     *                          zyklisch ausgeführt werdens soll.
     */
    public void setZyklischePrognose(final boolean zyklischePrognose) {
        this.zyklischePrognose = zyklischePrognose;
    }

    /**
     * Legt das Intervall fest in dem die Prognose überprüft werden soll.
     *
     * @param pruefIntervall das Prüfintervall.
     */
    public void setPruefIntervall(final Duration pruefIntervall) {
        this.pruefIntervall = pruefIntervall;
    }

    /**
     * Legt die Änderungsschwelle fest, bei der eine neue Prognoseganglinie gesendet
     * werden soll.
     *
     * @param schwelle die Schwelle.
     */
    public void setSchwelle(final int schwelle) {
        this.schwelle = schwelle;
    }

    /**
     * Legt das Intervall fest, in dem mindestens eine Prognoseganglinie versendet
     * werden soll.
     *
     * @param sendeIntervall das Intervall.
     */
    public void setSendeIntervall(final Duration sendeIntervall) {
        this.sendeIntervall = sendeIntervall;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        String s = getClass().getName() + "[";

        s += "messQuerschnitt=" + messQuerschnitt;
        s += ", prognoseZeitraum=" + prognoseZeitraum;
        s += ", nurLangfristigeAuswahl=" + nurLangfristigeAuswahl;
        s += ", ereignisTypen=" + ereignisTypen;
        s += ", zyklischePrognose=" + zyklischePrognose;
        s += ", pruefIntervall=" + pruefIntervall;
        s += ", schwelle=" + schwelle;
        s += ", sendeIntervall=" + sendeIntervall;

        return s + "]";
    }

}
