/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.portals;

import java.time.Duration;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

import de.bsvrz.iav.gllib.gllib.domain.Ereignis;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalender;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalenderAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalenderAntwort;
import de.bsvrz.iav.gllib.gllib.shared.Event;

/**
 * Kommunikator zum Versenden von Anfragen an den Ereigniskalender.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public final class DavEreignisKalender implements EreignisKalender {

    private DavEreignisKalenderAnfrageSender sender;
    private DavEreignisMengenObserver ereignisObserver;

    private Map<String, Consumer<EreignisKalenderAntwort>> versendeteAnfragen = Collections
            .synchronizedMap(new LinkedHashMap<>());

    private ScheduledExecutorService timeOutChecker;

    public DavEreignisKalender(DavObjektFactory objektFactory) {
        this.sender = new DavEreignisKalenderAnfrageSender(objektFactory.getDav());
        this.ereignisObserver = new DavEreignisMengenObserver(objektFactory);
        DavEreignisKalenderAntwortEmpfaenger empfaenger = new DavEreignisKalenderAntwortEmpfaenger(objektFactory);
        empfaenger.antwortEmpfangen().addHandler(a -> antwortEmpfangen(a));
        timeOutChecker = Executors.newScheduledThreadPool(10);
    }

    @Override
    public void kalenderAnfragen(EreignisKalenderAnfrage anfrage, Consumer<EreignisKalenderAntwort> antwortEmpfangen) {
        versendeteAnfragen.put(anfrage.getAbsenderZeichen(), antwortEmpfangen);
        sender.sendeAnfrage(anfrage);
    }

    @Override
    public void kalenderAnfragen(EreignisKalenderAnfrage anfrage, Consumer<EreignisKalenderAntwort> antwortEmpfangen,
            Duration timeout, Runnable timeoutErreicht) {
        kalenderAnfragen(anfrage, antwortEmpfangen);
        timeOutChecker.schedule(() -> {
            Consumer<EreignisKalenderAntwort> consumer = versendeteAnfragen.remove(anfrage.getAbsenderZeichen());
            if (consumer != null) {
                timeoutErreicht.run();
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

    @Override
    public Event<Boolean> ereignisMengeGeaendert() {
        return ereignisObserver.mengenAenderung();
    }

    private void antwortEmpfangen(EreignisKalenderAntwort a) {
        Consumer<EreignisKalenderAntwort> consumer = versendeteAnfragen.remove(a.getAbsenderZeichen());
        if (consumer != null) {
            consumer.accept(a);
        }
    }

    @Override
    public EreignisKalenderAntwort kalenderAnfragen(EreignisKalenderAnfrage anfrage, Duration timeout)
            throws TimeoutException {

        try {
            AtomicReference<EreignisKalenderAntwort> antwort = new AtomicReference<>(null);
            final Semaphore warteAufAnwort = new Semaphore(1);
            warteAufAnwort.acquire();
            kalenderAnfragen(anfrage, a -> {
                antwort.set(a);
                warteAufAnwort.release();
            });

            boolean antwortEingetroffen = warteAufAnwort.tryAcquire(timeout.toMillis(), TimeUnit.MILLISECONDS);
            if (antwortEingetroffen) {
                return antwort.get();
            }
        } catch (InterruptedException ignored) {
            // ignored
        }

        throw new TimeoutException("Ereigniskalender hat nicht in der erwarteten Zeit geantwortet");
    }

    @Override
    public Set<Ereignis> getEreignisse() {
        return ereignisObserver.getEreignisse();
    }
}
