/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.portals;

import java.time.Duration;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienAnfrager;
import de.bsvrz.iav.gllib.gllib.domain.comm.GanglinienPrognoseClient;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAntwort;
import de.bsvrz.iav.gllib.gllib.domain.comm.PrognoseAnfrage;

/**
 * Mit dieser Klasse können Applikation bequem Anfragen an die
 * Ganglinienprognose stellen, ohne sich mit dem Datenkatalog
 * auseinanderzusetzen müssen.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class DavGanglinienprognose implements GanglinienPrognoseClient {

    private static AtomicLong anfrageId = new AtomicLong(1);

    private DavPrognoseAnfrageSender sender;

    private Map<String, Consumer<GlProgAntwort>> versendeteAnfragen = Collections
            .synchronizedMap(new LinkedHashMap<>());

    private ScheduledExecutorService timeOutChecker;

    private DavPrognoseAntwortEmpfaenger empfaenger;

    private final GanglinienAnfrager anfrager;

    public DavGanglinienprognose(DavObjektFactory objektFactory) {
        sender = new DavPrognoseAnfrageSender(objektFactory.getDav());
        empfaenger = new DavPrognoseAntwortEmpfaenger(objektFactory.getDav(), objektFactory);
        empfaenger.antwortEmpfangen().addHandler(a -> antwortEmpfangen(a));
        timeOutChecker = Executors.newScheduledThreadPool(10);
        
        ClientApplication appObj = objektFactory.getDav().getLocalApplicationObject();
        anfrager = GanglinienAnfrager.of(appObj.getNameOrPidOrId(), appObj.getId());
    }

    private void antwortEmpfangen(GlProgAntwort a) {
        Consumer<GlProgAntwort> consumer = versendeteAnfragen.remove(a.getAbsenderZeichen());
        if (consumer != null) {
            consumer.accept(a);
        }
    }

    private void sendeAnfrage(GlProgAnfrage anfrage, Consumer<GlProgAntwort> antwortEmpfangen) {
        versendeteAnfragen.put(anfrage.getAbsenderZeichen(), antwortEmpfangen);
        sender.sendeAnfrage(anfrage);
    }

    @Override
    public void sendeAnfrage(String kennung, List<PrognoseAnfrage> anfragen, Consumer<GlProgAntwort> antwortEmpfangen) {
        GlProgAnfrage anfrage = createAnfrage(kennung, anfragen);
        sendeAnfrage(anfrage, antwortEmpfangen);
    }

    @Override
    public void sendeAnfrage(String kennung, List<PrognoseAnfrage> anfragen, Consumer<GlProgAntwort> antwortEmpfangen,
            Duration timeout, Runnable timeoutErreicht) {
        GlProgAnfrage anfrage = createAnfrage(kennung, anfragen);
        sendeAnfrage(anfrage, antwortEmpfangen);
        timeOutChecker.schedule(() -> {
            Consumer<GlProgAntwort> consumer = versendeteAnfragen.remove(anfrage.getAbsenderZeichen());
            if (consumer != null) {
                timeoutErreicht.run();
            }
        }, timeout.getSeconds(), TimeUnit.SECONDS);
    }

    @Override
    public GlProgAntwort sendeAnfrage(String kennung, List<PrognoseAnfrage> anfragen, Duration timeout)
            throws TimeoutException {
        GlProgAnfrage anfrage = createAnfrage(kennung, anfragen);
        try {
            AtomicReference<GlProgAntwort> antwort = new AtomicReference<>(null);
            final Semaphore warteAufAnwort = new Semaphore(1);
            warteAufAnwort.acquire();
            sendeAnfrage(anfrage, a -> {
                antwort.set(a);
                warteAufAnwort.release();
            });

            boolean antwortEingetroffen = warteAufAnwort.tryAcquire(timeout.toMillis(), TimeUnit.MILLISECONDS);
            if (antwortEingetroffen) {
                return antwort.get();
            }
        } catch (InterruptedException ignored) {
            // ignored
        }

        throw new TimeoutException("Die Ganglinienprognose hat nicht in der erwarteten Zeit geantwortet");
    }

    private GlProgAnfrage createAnfrage(String kennung, List<PrognoseAnfrage> anfragen) {
        String absenderZeichen = getAbsenderZeichen(kennung);
        return new GlProgAnfrage(anfrager, absenderZeichen, anfragen);
    }

    private String getAbsenderZeichen(String kennung) {
        if (kennung == null || kennung.isEmpty()) {
            return Long.toString(anfrageId.getAndIncrement());
        }
        return kennung;
    }
}
