/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.shared;

import java.time.LocalDateTime;
import java.util.Objects;

/**
 * Hilfsklasse zur Bearbeitung eines String in Cron-Syntax. Folgende
 * Syntax-Elemente werden unterstützt:
 * <p>
 * Der String besteht aus fünf Bereichen:
 * <ol>
 * <li>Minute: 0-59 oder * für jede Minute</li>
 * <li>Stunde: 0-23 oder * für jede Stunde</li>
 * <li>Tag des Monats: 1-31 oder * für jeden Tag</li>
 * <li>Monat: 1-12 oder * für jeden Monat</li>
 * <li>Wochentag: 0-7 oder * für jeden Wochentag (0 und 7 stehen beide für
 * Sonntag)</li>
 * </ol>
 * 
 * TODO Listen unterstützen. 
 * TODO Bereiche unterstützen. 
 * TODO Schrittweiten unterstützen. 
 * TODO Strings für Monat und Wochentag unterstützen.
 * 
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class CronPattern implements ValueMatcher<LocalDateTime> {

    public static final CronPattern NULL = new CronPattern("", null, null, null, null, null);

    private final String pattern;

    private final ValueMatcher<Integer> minuteMatcher;
    private final ValueMatcher<Integer> hourMatcher;
    private final ValueMatcher<Integer> dayOfMonthMatcher;
    private final ValueMatcher<Integer> monthMatcher;
    private final ValueMatcher<Integer> dayOfWeekMatcher;

    private CronPattern(String pattern, ValueMatcher<Integer> minuteMatcher, ValueMatcher<Integer> hourMatcher,
            ValueMatcher<Integer> dayOfMonthMatcher, ValueMatcher<Integer> monthMatcher,
            ValueMatcher<Integer> dayOfWeekMatcher) {
        this.pattern = pattern;
        this.minuteMatcher = minuteMatcher;
        this.hourMatcher = hourMatcher;
        this.dayOfMonthMatcher = dayOfMonthMatcher;
        this.monthMatcher = monthMatcher;
        this.dayOfWeekMatcher = dayOfWeekMatcher;
    }

    public static CronPattern parse(String pattern) {

        Objects.requireNonNull(pattern);
        String[] parts;

        parts = pattern.split(" |\t");
        if (parts.length != 5) {
            throw new IllegalArgumentException("Ungültiges cron-Pattern: '" + pattern + "'");
        }
        try {
            ValueMatcher<Integer> minuteMatcher;
            if (parts[0].equals("*")) {
                minuteMatcher = new ValueMatcher.AlwaysMatcher<>();
            } else {
                minuteMatcher = new ValueMatcher.CollectionValueMatcher<>(Integer.valueOf(parts[0]));
            }

            ValueMatcher<Integer> hourMatcher;
            if (parts[1].equals("*")) {
                hourMatcher = new ValueMatcher.AlwaysMatcher<>();
            } else {
                hourMatcher = new ValueMatcher.CollectionValueMatcher<>(Integer.valueOf(parts[1]));
            }

            ValueMatcher<Integer> dayOfMonthMatcher;
            if (parts[2].equals("*")) {
                dayOfMonthMatcher = new ValueMatcher.AlwaysMatcher<>();
            } else {
                dayOfMonthMatcher = new ValueMatcher.CollectionValueMatcher<>(Integer.valueOf(parts[2]));
            }

            ValueMatcher<Integer> monthMatcher;
            if (parts[3].equals("*")) {
                monthMatcher = new ValueMatcher.AlwaysMatcher<>();
            } else {
                // Java-Monate gehen bei 0 los
                monthMatcher = new ValueMatcher.CollectionValueMatcher<>(Integer.valueOf(parts[3]));
            }

            ValueMatcher<Integer> dayOfWeekMatcher;
            if (parts[4].equals("*")) {
                dayOfWeekMatcher = new ValueMatcher.AlwaysMatcher<>();
            } else {
                // Java-Wochentage gehen bei 1 mit Sonntag los
                int day;

                day = Integer.parseInt(parts[4]);
                if (day == 0) {
                    // Sonntag ist intern nur 7, nicht mehr auch 0
                    day = 7;
                }
                dayOfWeekMatcher = new ValueMatcher.CollectionValueMatcher<>(day);
            }

            return new CronPattern(pattern, minuteMatcher, hourMatcher, dayOfMonthMatcher, monthMatcher,
                    dayOfWeekMatcher);

        } catch (@SuppressWarnings("unused") final NumberFormatException ignored) {
            throw new IllegalArgumentException("Das Pattern '" + pattern
                    + "' entspricht nicht der unterstützten Cron-Syntax. Listen, Intervalle und Schrittweiten werden noch nicht unterstützt."
                    + pattern + "'");
        }

    }

    /**
     * Gibt den zugrundeliegenden String zurück.
     * 
     * @return der Original-String in Cron-Syntax.
     */
    public String getPattern() {
        return pattern;
    }

    /**
     * Prüft ob ein Zeitstampel mit dem Cron-Pattern matcht. Der Zeitstempel muss
     * mit den Angaben von Wochentag, Monat, Tag, Stunde und Minute übereinstimmen.
     * 
     * @param now ein beliebiger Zeitstempel.
     * @return {@code true}, wenn der Zeitstempel mit dem Cron-Pattern matcht.
     */
    @Override
    public boolean match(final LocalDateTime now) {

        if (isNull()) {
            return false;
        }

        int minute = now.getMinute();
        int hour = now.getHour();
        int dayOfMonth = now.getDayOfMonth();

        // Java-Monate gehen bei 0 los
        int month = now.getMonthValue();

        // Java-Wochentage gehen bei 1 mit Sonntag los
        int dayOfWeek = now.getDayOfWeek().getValue();

        boolean result = minuteMatcher.match(minute);
        result &= hourMatcher.match(hour);
        result &= dayOfMonthMatcher.match(dayOfMonth);
        result &= monthMatcher.match(month);
        result &= dayOfWeekMatcher.match(dayOfWeek);

        return result;
    }

    public boolean nonNull() {
        return !isNull();
    }

    public boolean isNull() {
        return getPattern().isEmpty();
    }

    @Override
    public String toString() {
        String s;

        s = getClass().getName() + "[";
        s += "pattern=" + getPattern();
        s += "]";
        return s;
    }

}
