/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.domain;

import java.text.DateFormat;
import java.util.Date;
import java.util.Objects;

import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;

/**
 * Repräsentiert ein Intervall für Zeitstempel als Millisekunden. Wird für
 * Zeitintervalle genutzt, die mit Zeitstempeln arbeiten.
 * 
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class Intervall implements Cloneable {

    private final long start;

    private final long end;

    /**
     * Konstruiert ein Zeitintervall mit dem angegebenen Grenzen.
     * 
     * @param start Start des Intervalls
     * @param end   Ende des Intervalls
     */
    private Intervall(final long start, final long end) {
        if (start > end) {
            throw new IllegalArgumentException("Das Ende des Intervalls darf nicht vor dessen Start liegen (start="
                    + start + ", ende=" + end + ").");
        }

        this.start = start;
        this.end = end;
    }

    public static Intervall of(long start, long ende) {
        return new Intervall(start, ende);
    }

    public static Intervall of(PrognoseIntervall prognoseIntervall) {
        return new Intervall(LocalDateTimeKonverter.toMillis(prognoseIntervall.getVon()),
                LocalDateTimeKonverter.toMillis(prognoseIntervall.getBis()));
    }

    @Override
    public Intervall clone() {
        try {
            return (Intervall) super.clone();
        } catch (final CloneNotSupportedException ex) {
            throw new IllegalStateException(ex);
        }
    }

    /**
     * Prüft ob ein anderes Intervall in diesem Intervall enhalten ist, inklusive
     * der Intervallgrenzen.
     * 
     * @param interval ein Intervall.
     * @return {@code true}, wenn das andere Intervall innerhalb dieses Intervalls
     *         liegt oder mit ihm identisch ist.
     */
    public boolean contains(final Intervall interval) {
        return contains(interval.start) && contains(interval.end);
    }

    /**
     * Prüft ob ein Wert im Intervall enhalten ist, inklusive der Intervallgrenzen.
     * 
     * @param value Ein Wert
     * @return {@code true}, wenn der Wert innerhalb des Intervalls oder auf einer
     *         der Intervallgrenzen liegt
     */
    public boolean contains(final long value) {
        return start <= value && value <= end;
    }

    /**
     * Zwei Intervalle sind gleich, wenn sie den selben Start- und Endwert besitzen.
     * 
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof Intervall) {
            final Intervall intervall = (Intervall) obj;
            boolean result = Objects.equals(start, intervall.start);
            return result && Objects.equals(end, intervall.end);
        }

        return false;
    }

    /**
     * Gibt das Ende des Intervalls zurück.
     * 
     * @return Zeitstempel
     */
    public long getEnd() {
        return end;
    }

    /**
     * Gibt den Anfang des Intervalls zurück.
     * 
     * @return Zeitstempel
     */
    public long getStart() {
        return start;
    }

    @Override
    public int hashCode() {
        return Objects.hash(start, end);
    }

    /**
     * Prüft ob sich zwei Intervalle schneiden. Zwei Intervalle schneiden sich, wenn
     * sie mindestens einen Punkt gemeinsam haben (exklusive den Intervallgrenzen)
     * und ein Intervall nicht das andere enthält.
     * 
     * @param interval Ein anderes Intervall
     * @return {@code true}, wenn sich dieses Intervall mit dem anderen schneidet
     */
    public boolean intersect(final Intervall interval) {
        return (!contains(interval) && !interval.contains(this)) && (interval.isWithin(start) || interval.isWithin(end)
                || isWithin(interval.start) || isWithin(interval.end));
    }

    /**
     * Prüft ob sich zwei Intervalle berühren. Zwei Intervalle berühren sich, wenn
     * sie mindestens einen Punkt gemeinsam haben (inklusive den Intervallgrenzen).
     * 
     * @param interval ein anderes Interval.
     * @return {@code true}, wenn dieses Intervall das andere berührt.
     */
    public boolean touch(final Intervall interval) {
        return interval.contains(start) || interval.contains(end) || contains(interval.start) || contains(interval.end);
    }

    /**
     * Prüft ob sich ein anderes Intervall innerhalb der Intervallgrenzen dieses
     * Intervalls befindet.
     * 
     * @param interval ein Intervall.
     * @return {@code true}, wenn das andere Intervall innerhalb dieses Intervalls
     *         liegt, aber nicht mit ihm identisch ist.
     */
    public boolean isWithin(final Intervall interval) {
        return isWithin(interval.start) && isWithin(interval.end);
    }

    /**
     * Prüft ob sich ein Wert innerhalb der Intervallgrenzen befindet.
     * 
     * @param value Ein Wert
     * @return {@code true}, wenn der Wert innerhalb des Intervalls, aber nicht auf
     *         einer der Intervallgrenzen liegt
     */
    public boolean isWithin(final long value) {
        return start < value && value < end;
    }

    /**
     * Gibt das Intervall als für den Menschen lesbaren Text zurück. Handelt es sich
     * um ein Zeitintervall, werden die Zeitstempel als Datum und Uhrzeit
     * formatiert.
     * 
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return "[" + DateFormat.getDateTimeInstance().format(new Date(start)) + ", "
                + DateFormat.getDateTimeInstance().format(new Date(end)) + "]";
    }

    public long getLength() {
        return end - start;
    }
}
