/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */
package de.bsvrz.iav.gllib.gllib.portals;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.concurrent.TimeUnit;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.iav.gllib.gllib.domain.Archiv;
import de.bsvrz.iav.gllib.gllib.domain.BetriebsMeldungsSender;
import de.bsvrz.iav.gllib.gllib.domain.DoubleGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implementierung der Schnittstelle {@link Archiv} die Daten über das Archiv
 * eines Datenverteiler-Systems bereitstellt.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavArchiv implements Archiv {

    private static final Debug LOGGER = Debug.getLogger();

    private ClientDavInterface dav;

    private BetriebsMeldungsSender betriebsmeldungsSender;

    public DavArchiv(ClientDavInterface dav, BetriebsMeldungsSender betriebsmeldungsSender) {
        this.dav = dav;
        this.betriebsmeldungsSender = betriebsmeldungsSender;
    }

    @Override
    public MqGanglinie bestimmeArchivGanglinie(String aspect, MessQuerschnitt messQuerschnitt, Intervall zeitraum) {
        final ArchiveRequestManager archiv;
        final ArchiveDataKindCombination dataKinds;
        final DataDescription dbs;
        final ArchiveTimeSpecification timeSpec;
        final ArchiveDataSpecification spec;
        final ArchiveDataQueryResult antwort;
        final AttributeGroup atg;
        final Aspect asp;
        final DoubleGanglinie gQKfz, gQLkw, gVPkw, gVLkw;

        archiv = dav.getArchive();
        if (!archiv.isArchiveAvailable()) {
            String message = MessageFormat.format("{0}: Das Archiv ist nicht verfügbar.", messQuerschnitt);
            LOGGER.error(() -> message);
            betriebsmeldungsSender.fehler(messQuerschnitt, message);
            return new MqGanglinie(messQuerschnitt);
        }

        // Anfrage zusammenstellen
        atg = dav.getDataModel().getAttributeGroup("atg.verkehrsDatenKurzZeitMq");
        asp = dav.getDataModel().getAspect(aspect);

        dbs = new DataDescription(atg, asp);
        timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, zeitraum.getStart() - 1,
                zeitraum.getEnd() + TimeUnit.DAYS.toMillis(1));
        dataKinds = new ArchiveDataKindCombination(ArchiveDataKind.ONLINE);
        spec = new ArchiveDataSpecification(timeSpec, dataKinds, ArchiveOrder.BY_DATA_TIME, ArchiveRequestOption.NORMAL,
                dbs, dav.getDataModel().getObject(messQuerschnitt.getPid()));

        // Anfrage verschicken und auf Antwort warten
        antwort = archiv.request(ArchiveQueryPriority.MEDIUM, spec);

        gQKfz = new DoubleGanglinie();
        gQLkw = new DoubleGanglinie();
        gVPkw = new DoubleGanglinie();
        gVLkw = new DoubleGanglinie();
        try {
            final ArchiveDataStream strom;
            ArchiveData archivdaten;

            assert antwort.getStreams().length == 1 : "Es darf nur einen Archivstream geben.";

            // Daten auslesen und Ganglinie daraus erstellen
            strom = antwort.getStreams()[0];
            while ((archivdaten = strom.take()) != null) {
                final DataState status;

                status = archivdaten.getDataType();
                if (!status.equals(DataState.DATA) && !status.equals(DataState.NO_DATA)
                        && !status.equals(DataState.NO_SOURCE)) {
                    // Statuswerte des Archiv werden nicht in Ganglinie
                    // übernommen
                    continue;
                }

                final Data daten;
                final long zeitstempel;
                final Double qKfz;
                final Double qLkw;
                final Double vLkw;
                final Double vPkw;

                zeitstempel = archivdaten.getDataTime();
                daten = archivdaten.getData();
                if (daten == null) {
                    // Definitionslücke
                    qKfz = null;
                    qLkw = null;
                    vLkw = null;
                    vPkw = null;
                } else {
                    if (daten.getItem("QKfz").getScaledValue("Wert") //$NON-NLS-2$
                            .isState()) {
                        // Definitionslücke
                        qKfz = null;
                    } else {
                        qKfz = daten.getItem("QKfz").getScaledValue("Wert") //$NON-NLS-2$
                                .doubleValue();
                    }
                    if (daten.getItem("QLkw").getScaledValue("Wert") //$NON-NLS-2$
                            .isState()) {
                        // Definitionslücke
                        qLkw = null;
                    } else {
                        qLkw = daten.getItem("QLkw").getScaledValue("Wert") //$NON-NLS-2$
                                .doubleValue();
                    }
                    if (daten.getItem("VLkw").getScaledValue("Wert") //$NON-NLS-2$
                            .isState()) {
                        // Definitionslücke
                        vLkw = null;
                    } else {
                        vLkw = daten.getItem("VLkw").getScaledValue("Wert") //$NON-NLS-2$
                                .doubleValue();
                    }
                    if (daten.getItem("VPkw").getScaledValue("Wert") //$NON-NLS-2$
                            .isState()) {
                        // Definitionslücke
                        vPkw = null;
                    } else {
                        vPkw = daten.getItem("VPkw").getScaledValue("Wert") //$NON-NLS-2$
                                .doubleValue();
                    }
                }
                gQKfz.aktualisierePunkt(zeitstempel, qKfz);
                gQLkw.aktualisierePunkt(zeitstempel, qLkw);
                gVPkw.aktualisierePunkt(zeitstempel, vPkw);
                gVLkw.aktualisierePunkt(zeitstempel, vLkw);
            }

            MqGanglinie ganglinie = new MqGanglinie(messQuerschnitt);
            ganglinie.aktualisiereMesswerte(GanglinienMQOperationen.zusammenfuehren(gQKfz, gQLkw, gVPkw, gVLkw));
            if (ganglinie.getStuetzStellenAnzahl() >= 2) {
                GanglinienMQOperationen.verschiebeUmHalbesIntervall(ganglinie);
            }
            if (ganglinie.getStuetzStellenAnzahl() > 0) {
                GanglinienMQOperationen.auschneiden(ganglinie, zeitraum);

                LOGGER.fine(() -> MessageFormat.format("{0}: Analyseganglinie aus Archivdaten: {1}", messQuerschnitt,
                        GanglinienMQOperationen.formatierterText(ganglinie)));

                return ganglinie;
            }
            LOGGER.info(() -> MessageFormat.format("{0}: Im Lernzeitraum wurden keine Kurzzeitdaten archiviert.",
                    messQuerschnitt));
        } catch (final IllegalStateException | InterruptedException | IOException ex) {
            String message = MessageFormat.format("{0}: Bei der Archivanfrage ist ein Fehler aufgetreten: {1}",
                    messQuerschnitt, ex.getLocalizedMessage());
            LOGGER.error(() -> message);
            betriebsmeldungsSender.fehler(messQuerschnitt, message);
        }

        return new MqGanglinie(messQuerschnitt);

    }

    @Override
    public boolean isArchiveAvailable() {
        return dav.getArchive().isArchiveAvailable();
    }
}
