/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.5 Funktionen Ganglinie
 * Copyright (C) 2011-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.portals;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.Ereignis;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTypLernParameter;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienAnfrager;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.ObjektFactory;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;

/**
 * Factory zum Erstellen von Objekten des Domain-Modells auf Basis von
 * Dav-Konfigurationsobjekten.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavObjektFactory implements ObjektFactory {

    private ClientDavInterface dav;

    private final Map<String, String> mqStrassensegmentPidCache = new LinkedHashMap<>();

    public DavObjektFactory(ClientDavInterface dav) {
        this.dav = dav;
    }

    public GanglinienAnfrager createAnfrager(long id) {
        SystemObject so = dav.getDataModel().getObject(id);
        if (so.isValid() && so.isOfType("typ.applikation")) {
            return GanglinienAnfrager.of(so.getName(), so.getId());
        }
        return GanglinienAnfrager.of(Long.toString(id), id);
    }

    public Ereignis createEreignis(String pid) {
        SystemObject so = dav.getDataModel().getObject(pid);
        if (so.isOfType("typ.ereignis")) {

            Data ereignisDaten = so
                    .getConfigurationData(dav.getDataModel().getAttributeGroup("atg.ereignisEigenschaften"));
            SystemObject etObject = ereignisDaten.getReferenceValue("EreignisTypReferenz").getSystemObject();

            EreignisTyp ereignisTyp = createEreignisTyp(etObject);
            Ereignis ereignis = Ereignis.of(so.getPid(), ereignisTyp);

            return ereignis;
        }
        return null;
    }

    private EreignisTyp createEreignisTyp(SystemObject so) {

        if (so == null || !so.isValid() || !so.isOfType("typ.ereignisTyp")) {
            return null;
        }

        EreignisTyp et = EreignisTyp.of(so.getPid());

        ResultData ereignisTypDaten = dav.getData(so,
                new DataDescription(dav.getDataModel().getAttributeGroup("atg.ereignisTypParameter"),
                        dav.getDataModel().getAspect("asp.parameterSoll")),
                0);
        if (ereignisTypDaten.hasData()) {
            et.setPrioritaet(ereignisTypDaten.getData().getUnscaledValue("EreignisTypPriorität").longValue());
        }

        ResultData ganglinienLernenDaten = dav.getData(so,
                new DataDescription(
                        dav.getDataModel().getAttributeGroup("atg.ganglinienModellAutomatischesLernenEreignis"),
                        dav.getDataModel().getAspect("asp.parameterSoll")),
                0);
        if (ganglinienLernenDaten.hasData()) {
            Data data = ganglinienLernenDaten.getData();
            EreignisTypLernParameter lernParameter = new EreignisTypLernParameter();

            Array array = data.getArray("AlgAusschlussliste");
            for (int idx = 0; idx < array.getLength(); idx++) {
                EreignisTyp typ = createEreignisTyp(array.getReferenceValue(idx).getSystemObject());
                if (typ != null) {
                    lernParameter.addAusschluss(typ);
                }
            }

            array = data.getArray("AlgBezugsereignistypen");
            for (int idx = 0; idx < array.getLength(); idx++) {
                EreignisTyp typ = createEreignisTyp(array.getReferenceValue(idx).getSystemObject());
                if (typ != null) {
                    lernParameter.addBezug(typ);
                }
            }

            lernParameter.setDarstellungsVerfahren(
                    ApproximationTyp.with(data.getUnscaledValue("AlgDarstellungsverfahren").intValue()));
            lernParameter.setGanglinieTyp(GanglinieTyp.with(data.getUnscaledValue("AlgGanglinienTyp").intValue()));
            lernParameter.setMatchingIntervallNach(
                    Duration.ofSeconds(data.getUnscaledValue("AlgMatchingIntervallNach").longValue()));
            lernParameter.setMatchingIntervallVor(
                    Duration.ofSeconds(data.getUnscaledValue("AlgMatchingIntervallVor").longValue()));
            lernParameter.setMatchingSchrittweite(
                    Duration.ofSeconds(data.getUnscaledValue("AlgMatchingSchrittweite").longValue()));
            lernParameter.setMaxGanglinien(data.getUnscaledValue("AlgMaxGanglinien").longValue());
            lernParameter.setMaximalerAbstand(data.getUnscaledValue("AlgMaxAbstand").intValue());
            lernParameter.setMaximalerMatchingFehler(data.getUnscaledValue("AlgMaxMatchingFehler").intValue());
            lernParameter.setMaximalerWichtungsFaktor(data.getUnscaledValue("AlgMaxWichtungsfaktor").intValue());
            lernParameter.setVergleichsSchrittweite(
                    Duration.ofSeconds(data.getUnscaledValue("AlgVergleichsSchrittweite").longValue()));

            et.setLernParameter(lernParameter);
        }

        return et;
    }

    @Override
    public EreignisTyp createEreignisTyp(String pid) {
        return createEreignisTyp(dav.getDataModel().getObject(pid));
    }

    public EreignisTyp createEreignisTyp(long id) {
        return createEreignisTyp(dav.getDataModel().getObject(id));
    }

    public MessQuerschnitt createMessQuerschnitt(SystemObject so) {

        if (!so.isOfType("typ.messQuerschnittAllgemein")) {
            return null;
        }

        MessQuerschnitt messQuerschnitt = MessQuerschnitt.of(so.getPid());
        String strassenSegmentPid = mqStrassensegmentPidCache.get(so.getPid());
        if (strassenSegmentPid == null) {

            AttributeGroup atg = dav.getDataModel().getAttributeGroup("atg.punktLiegtAufLinienObjekt");
            Data data = so.getConfigurationData(atg);
            if (data != null) {
                SystemObject referenz = data.getReferenceValue("LinienReferenz").getSystemObject();
                if (referenz != null) {
                    if (referenz.isOfType("typ.äußeresStraßenSegment")) {
                        strassenSegmentPid = referenz.getPid();
                        mqStrassensegmentPidCache.put(messQuerschnitt.getPid(), strassenSegmentPid);
                    } else if (referenz.isOfType("typ.straßenTeilSegment")) {
                        for (SystemObject ss : dav.getDataModel().getType("typ.äußeresStraßenSegment").getElements()) {
                            AttributeGroup ssAtg = dav.getDataModel()
                                    .getAttributeGroup("atg.punktLiegtAufLinienObjekt");
                            Data ssData = so.getConfigurationData(ssAtg);
                            Array array = ssData.getArray("LinienReferenz");
                            for (int idx = 0; idx < array.getLength(); idx++) {
                                if (array.getReferenceValue(idx).getSystemObject().equals(referenz)) {
                                    strassenSegmentPid = ss.getPid();
                                    mqStrassensegmentPidCache.put(messQuerschnitt.getPid(), strassenSegmentPid);
                                    break;
                                }
                            }
                            if (strassenSegmentPid != null) {
                                break;
                            }
                        }
                    }
                }
            }
        }

        messQuerschnitt.setStrassenSegmentPid(strassenSegmentPid);
        return messQuerschnitt;
    }

    @Override
    public MessQuerschnitt createMessQuerschnitt(String pid) {
        SystemObject so = dav.getDataModel().getObject(pid);
        if (so == null) {
            return null;
        }
        return createMessQuerschnitt(so);
    }

    public ClientDavInterface getDav() {
        return dav;
    }

    public SystemObject getGanglinienSpeicherObjekt(String dbSpeicherPid) {

        SystemObject speicherObjekt = null;

        if (dbSpeicherPid == null || dbSpeicherPid.isEmpty()) {
            List<SystemObject> speicherObjects = dav.getDataModel().getType("typ.ganglinienSpeicher").getElements();
            if (speicherObjects.isEmpty()) {
                throw new IllegalStateException("Es konnte kein Objekt vom Typ 'GanglinienSpeicher' ermittelt werden.");
            }
            if (speicherObjects.size() > 1) {
                throw new IllegalStateException(
                        "Es sind mehrere GanglinienSpeicher verfügbar: " + speicherObjects.toString());
            }
            speicherObjekt = speicherObjects.get(0);
        } else {
            speicherObjekt = dav.getDataModel().getObject(dbSpeicherPid);
            if (speicherObjekt == null || !speicherObjekt.isOfType("typ.ganglinienSpeicher")) {
                throw new IllegalArgumentException(
                        "Die PID '" + dbSpeicherPid + "' beschreibt keinen Ganglinienspeicher");
            }
        }

        return speicherObjekt;
    }

    @Override
    public LocalDateTime getTime() {
        return LocalDateTimeKonverter.fromMillis(dav.getTime());
    }
}
